/*
 *  connections.c : Manage connected sockets and the incoming data.
 *                  This file is part of the FreeLCD package.
 *
 *  $Id: connections.c,v 1.8 2004/06/20 13:02:08 unicorn Exp $
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright (c) 2002, 2003, Jeroen van den Berg <unicorn@hippie.nu>
 */

/** \file connections.c */

#include <errno.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#else
# ifdef HAVE_SYS_UNISTD_H
#  include <sys/unistd.h>
# endif
#endif

#if HAVE_ASSERT_H
# include <assert.h>
#else
# define assert(FOO)
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <glib.h>

#include "common/socket.h"
#include "common/debug.h"
#include "connections.h"
#include "scheduler.h"


/** This struct keeps the state information of a connection to a client. */
typedef struct
{
  /** Socket handle. */
  int        sock;

  GIOChannel *source;
  /** Address of the other side of the connection. */
  in_addr_t  originator;
  /** The XML parser context associated with this connection. */
  void       *xml_parser_context;
  /** A list of screens that the client has defined. It contains elements
   * of type const char *. */
  GSList     *registered_screens;
}
connection;

/** The list of active connections. */
static GSList *connectionlist = NULL;

/*------------------------------------------------------------------------*/
/* XML dictionaries  */

typedef enum
{
  SCREEN = 1, DATA, COMMAND, IMPORTANCE, REPEAT, FREQUENCY, LAYOUT
}
tag_t;

static tag_t conn_tagindex[] =
{
  SCREEN, DATA, COMMAND, IMPORTANCE, REPEAT, FREQUENCY, LAYOUT
};

static dict_pair conn_tags_d[] =
{
  { "command",    conn_tagindex + 2 },
  { "data",       conn_tagindex + 1 },
  { "frequency",  conn_tagindex + 5 },
  { "layout",     conn_tagindex + 6 },
  { "importance", conn_tagindex + 3 },
  { "repeat",     conn_tagindex + 4 },
  { "screen",     conn_tagindex + 0 }
};

static dictionary conn_tags = { conn_tags_d, 
                               sizeof (conn_tags_d) / sizeof (dict_pair) };


typedef enum
{
  NAME = 1
}
attr_t;

static attr_t conn_attrindex[] = 
{
  NAME
};

static dict_pair conn_attrs_d[] =
{
  { "name",     &conn_attrindex[0] }
};

static dictionary conn_attrs = { conn_attrs_d,
                                 sizeof (conn_attrs_d) / sizeof (dict_pair) };




/*----------------------------------------------------- _string_compare --*/

/** An internal function, used as an extra argument to g_slist_find_custom().
 * \param _str1 First string to compare.
 * \param _str2 Second string to compare.
 * \return 0 if the strings are not equal, any other value if they are.
 */
static gint
_string_compare (gconstpointer _str1, gconstpointer _str2)
{
  const char *str1 = (const char *)_str1;
  const char *str2 = (const char *)_str2;

  assert (str1 != NULL);
  assert (str2 != NULL);

  return strcmp (str1, str2);
}

/*---------------------------------------------------------- _elem_free --*/
static void
_elem_free (gpointer data, gpointer user_data)
{
  (void)user_data;
  g_free (data);
}

/*------------------------------------------------------ _handle_screen --*/

/** Handle screen definition data. If the screen was not registered yet, it
 * is added to the list. If it was registered, its settings are updated.
 * \param info The connection status associated with the client that sent
 *             the screen definition.
 * \param data Screen definition.
 */
static void
_handle_screen (connection *info, xml_node *data)
{
  sch_importance importance = IMP_NORMAL;
  int            repeat     = 10;
  const char*    name       = xmlt_get_attrib (data, NAME);
  xml_node       *found_lay = xmlt_find (data, NULL, LAYOUT);
  xml_node       *found_imp = xmlt_find (data, NULL, IMPORTANCE);
  xml_node       *found_rep = xmlt_find (data, NULL, REPEAT);
  
  if (found_lay == NULL)
    {
      /* This screen has no layout, and therefore no use. 
       * FIXME: Send an error. */
      debug ("no layout was found");

      return;
    }
  
  if (name == NULL)
    {
      /* Screens must have a name. FIXME: Error message. */
      debug ("screen has no name");

      return;
    }
  
  if (found_imp != NULL)
    {
      const char* importance_str = xmlt_get_first_cdata (found_imp);
  
      if (importance_str != NULL)
        {
          if (!strcmp (importance_str, "background"))
            importance = IMP_BACKGROUND;
  
          else if (!strcmp (importance_str, "normal"))
            importance = IMP_NORMAL;
  
          else if (!strcmp (importance_str, "notification"))
            importance = IMP_NOTIFICATION;
  
          else if (!strcmp (importance_str, "alert"))
            importance = IMP_ALERT;

          else
            {
              debug_2 ("%s is not a valid screen importance", importance_str);
              return;
            }
        }
    }
  
  if (found_rep != NULL)
    {
      const char* repeat_str = xmlt_get_first_cdata (found_rep);
  
      if (repeat_str != NULL)
        {
          if (!strcasecmp (repeat_str, "infinite"))
            repeat = -1;
  
          else
            {
              long tmp = strtol (repeat_str, 0, 10);
  
              if (errno == ERANGE || errno == EINVAL)
                {
                  /* FIXME: send an error message. */
                  return;
                }
                  
              if (tmp < 1 || tmp > 65536)
                {
                  /* FIXME: send an out of range error message. */
                  return;
                }
  
              repeat = tmp;
            }
        }
    }
  
  if (   g_slist_find_custom (info->registered_screens, name, _string_compare)
      == NULL)
    {
      debug_2 ("adding screen %s to list of registered screens", name);
      info->registered_screens = 
          g_slist_append (info->registered_screens, (void *)g_strdup (name));
    }

  sch_add_screen ((void *)info, name, importance, repeat, found_lay);
}

/*-------------------------------------------------------- _handle_data --*/
static void
_handle_data (connection *info, xml_node *data)
{
  const char *name = xmlt_get_attrib (data, NAME);

  if (name == NULL)
    {
      /* FIXME: send an error message. */
      return;
    }
  
  if (g_slist_find_custom (info->registered_screens, name, _string_compare)
      == NULL)
    {
      /* No screen has been registered for this data set yet.
       * FIXME: send an error message. */
      debug_2 ("no screen named %s registered yet.", name);
      return;
    }

  debug_2 ("updating data for %s.", name);
  sch_update_data ((void *)info, name, data);
}

/*----------------------------------------------------- _handle_command --*/
static void
_handle_command (connection *info, xml_node *data)
{
  /* FIXME: implement */
  (void)info;
  (void)data;
}

/*------------------------------------------------------- _xml_callback --*/
static void 
_xml_callback (void *_info, xml_node *data)
{
  connection *info = (connection *)_info;
  
  /* Find out what kind of message we received by looking at the root tag. */
  switch (data->tag)
    {
    case SCREEN:
      _handle_screen (info, data);
      break;
    
    case DATA:
      _handle_data (info, data);
      break;

    case COMMAND:
      _handle_command (info, data);
      break;

    default:
      /* FIXME: Send back an error message or something. */
      break;
    }

  xmlt_free_document (data);
}

/*-------------------------------------------------------- _xml_compare --*/
static gint
_source_compare (gconstpointer _conn, gconstpointer _source)
{
  const connection *conn = (const connection *)_conn;
  const GIOChannel *source = (const GIOChannel *)_source;

  return conn->source != source;
}

/*--------------------------------------------------- _connection_close --*/
static void
_connection_close (gpointer _conn, gpointer user_data)
{
  connection *conn = (connection *)_conn;
  (void)user_data;
  
  assert (conn != NULL);

  g_io_channel_unref (conn->source);
  g_slist_foreach (conn->registered_screens, _elem_free, NULL);
  g_slist_free (conn->registered_screens);
  xmlt_free_context (conn->xml_parser_context);
  g_free (conn);
}

/*----------------------------------------------------------- conn_open --*/
gboolean
conn_open (GIOChannel *source, GIOCondition condition, gpointer data)
{
  connection          *new_conn = g_new (connection, 1);
  GIOChannel          *new_source;
  gint                new_sock;
  struct sockaddr_in  s_in;
  socklen_t           sock_len = 0;

  assert (new_conn != NULL);
  debug ("conn open");
  
  if (condition & G_IO_IN)
    {
      new_sock = accept (g_io_channel_unix_get_fd (source), 
                         (struct sockaddr *)&s_in, &sock_len);
      
      if (new_sock < 0)
        return FALSE;
      
      new_source = g_io_channel_unix_new (new_sock);
      
      new_conn->sock = new_sock;
      new_conn->source = new_source;
      new_conn->originator = s_in.sin_addr.s_addr;
      new_conn->xml_parser_context = 
        xmlt_create_context (_xml_callback, new_conn, &conn_tags, &conn_attrs);

      new_conn->registered_screens = NULL;
      connectionlist = g_slist_append (connectionlist, new_conn);

      g_io_add_watch (new_source, G_IO_IN, conn_data, NULL);
      g_io_add_watch (new_source, G_IO_HUP, conn_close, NULL);
    }

  return TRUE;
}

/*----------------------------------------------------------- conn_data --*/
gboolean
conn_data (GIOChannel *source, GIOCondition condition, gpointer data)
{
  GSList     *found = g_slist_find_custom (connectionlist, source, 
                                           _source_compare);
  connection *conn = (connection *) found->data;
  gchar      buf[1024];
  guint      bytes_read;
  GIOError   error = G_IO_ERROR_NONE;

  if ((condition & G_IO_HUP) > 0)
    {
      debug ("conn_data : connection closed");
      return conn_close (source, condition, data);
    }

  while (1)
    {
      error = g_io_channel_read (source, buf, sizeof (buf), &bytes_read);
      if (error == G_IO_ERROR_NONE)
        {
          if (bytes_read == 0)
            {
              debug ("conn_data : empty read");
              return FALSE;
            }

          buf[bytes_read] = '\0';
          debug_3 ("conn_data : %i bytes read\n%s\n", bytes_read, buf);

          if (!xmlt_parse (conn->xml_parser_context, buf, bytes_read))
            xmlt_reset_context (conn->xml_parser_context);

          break;
        }

      else if (error != G_IO_ERROR_AGAIN)
        break;
    }
  
  return TRUE;
}

/*---------------------------------------------------------- conn_close --*/
gboolean
conn_close (GIOChannel *source, GIOCondition condition, gpointer data)
{
  debug ("conn close");

  _connection_close (source, NULL);
  g_io_channel_unref (source);

  return TRUE;
}

/*---------------------------------------------------------- conn_reply --*/
int
conn_reply (void *originator, const char *data, size_t len)
{
/*
  connection *info = (connection *)originator;
  if (socket_send (info->sock, data, len) == -1)
    {
      GSList *found = g_slist_find (connectionlist, &info);

      if (found != NULL)
        {
          connectionlist = g_slist_remove_link (connectionlist, found);
          g_slist_free_1 (found);
          g_slist_free (info->registered_screens);
          xmlt_free_context (info->xml_parser_context);
          g_free (info);
        }

      return 0;
    }
*/
  return 1;
}

/*------------------------------------------------------- conn_close_all --*/
void conn_close_all ()
{
  g_slist_foreach (connectionlist, _connection_close, NULL);
  g_slist_free (connectionlist);
  connectionlist = NULL;
}
